/*
 * Copyright 2017 Steve Lhomme
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110-1301, USA
 */

import "dxgi1_4.idl";

typedef enum DXGI_HDR_METADATA_TYPE
{
    DXGI_HDR_METADATA_TYPE_NONE    = 0x00,
    DXGI_HDR_METADATA_TYPE_HDR10   = 0x01,
} DXGI_HDR_METADATA_TYPE;

typedef struct DXGI_HDR_METADATA_HDR10
{
    UINT16 RedPrimary[2];
    UINT16 GreenPrimary[2];
    UINT16 BluePrimary[2];
    UINT16 WhitePoint[2];
    UINT   MaxMasteringLuminance;
    UINT   MinMasteringLuminance;
    UINT16 MaxContentLightLevel;
    UINT16 MaxFrameAverageLightLevel;
} DXGI_HDR_METADATA_HDR10;

[
    object,
    local,
    uuid(3d585d5a-bd4a-489e-b1f4-3dbcb6452ffb)
]
interface IDXGISwapChain4 : IDXGISwapChain3
{
    HRESULT SetHDRMetaData(
        [in] DXGI_HDR_METADATA_TYPE Type,
        [in] UINT Size,
        [in] void* pMetaData
    );
};

[
    object,
    local,
    uuid(80A07424-AB52-42EB-833C-0C42FD282D98)
]
interface IDXGIOutput5 : IDXGIOutput4
{
    HRESULT DuplicateOutput1(
        [in]        IUnknown               *pDevice,
                    UINT                   Flags,
        [in]        UINT                   SupportedFormatsCount,
        [in]  const DXGI_FORMAT            *pSupportedFormats,
        [out]       IDXGIOutputDuplication **ppOutputDuplication
    );
}
